@@ -1,7 +1,7 @@
 /*
  * number.h - Public API for Scheme numbers
  *
- *   Copyright (c) 2000-2021  Shiro Kawai  <shiro@acm.org>
+ *   Copyright (c) 2000-2022  Shiro Kawai  <shiro@acm.org>
  *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -244,6 +244,23 @@ SCM_EXTERN uint64_t Scm_GetIntegerU64Clamp(ScmObj obj, int clamp, int *oor);
 #define Scm_MakeIntegerFromUI(x) Scm_MakeIntegerU(x)
 #define Scm_GetUInteger(x)       Scm_GetIntegerU(x)
 
+/* System-dependent integral types */
+SCM_EXTERN int       Scm_IntegerFitsSizeP(ScmObj);
+SCM_EXTERN size_t    Scm_IntegerToSize(ScmObj);
+SCM_EXTERN ScmObj    Scm_SizeToInteger(size_t);
+SCM_EXTERN int       Scm_IntegerFitsSsizeP(ScmObj);
+SCM_EXTERN ssize_t   Scm_IntegerToSsize(ScmObj);
+SCM_EXTERN ScmObj    Scm_SsizeToInteger(ssize_t);
+SCM_EXTERN int       Scm_IntegerFitsPtrdiffP(ScmObj);
+SCM_EXTERN ptrdiff_t Scm_IntegerToPtrdiffP(ScmObj);
+SCM_EXTERN ScmObj    Scm_PtrdiffToInteger(ssize_t);
+SCM_EXTERN int       Scm_IntegerFitsOffsetP(ScmObj);
+SCM_EXTERN off_t     Scm_IntegerToOffset(ScmObj);
+SCM_EXTERN ScmObj    Scm_OffsetToInteger(off_t);
+SCM_EXTERN int       Scm_IntegerFitsIntptrP(ScmObj);
+SCM_EXTERN intptr_t  Scm_IntegerToIntptr(ScmObj);
+SCM_EXTERN ScmObj    Scm_IntptrToInteger(intptr_t);
+
 SCM_EXTERN ScmObj Scm_MakeRational(ScmObj numer, ScmObj denom);
 SCM_EXTERN ScmObj Scm_MakeRatnum(ScmObj numer, ScmObj denom);
 SCM_EXTERN ScmObj Scm_ReduceRational(ScmObj rational);
