@@ -1,7 +1,7 @@
 /*
  * gauche/port.h - Port API
  *
- *   Copyright (c) 2000-2019  Shiro Kawai  <shiro@acm.org>
+ *   Copyright (c) 2000-2022  Shiro Kawai  <shiro@acm.org>
  *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -88,20 +88,17 @@ typedef struct ScmPortBufferRec {
     int  (*filenum)(ScmPort *p);
     off_t (*seeker)(ScmPort *p, off_t offset, int whence);
     void *data;
+
+    /* The following fields are added in 0.9.10.  When you pass ScmPortBuffer
+       to the constructor, those fields are looked at only when you specify
+       SCM_PORT_WITH_POSITION.  They are needed to fully support srfi-192-style
+       port positioning.   See the Port Positioning Interface below. */
+    ScmObj (*getpos)(ScmPort *p);
+    void   (*setpos)(ScmPort *p, ScmObj pos);
+    u_long flags;               /* reserved */
 } ScmPortBuffer;
 
-/* For input buffered port, returns the size of room that can be filled
-   by the filler */
-#define SCM_PORT_BUFFER_ROOM(p) \
-    (int)((p)->src.buf.buffer+(p)->src.buf.size-(p)->src.buf.end)
-
-/* For output buffered port, returns the size of available data that can
-   be flushed by the flusher */
-#define SCM_PORT_BUFFER_AVAIL(p) \
-    (int)((p)->src.buf.current-(p)->src.buf.buffer)
-
 /* The function table of procedural port. */
-
 typedef struct ScmPortVTableRec {
     int     (*Getb)(ScmPort *p);
     int     (*Getc)(ScmPort *p);
@@ -115,74 +112,104 @@ typedef struct ScmPortVTableRec {
     void    (*Close)(ScmPort *p);
     off_t   (*Seek)(ScmPort *p, off_t off, int whence);
     void    *data;
+
+    /* The following fields are added in 0.9.10.  When you pass ScmPortBuffer
+       to the constructor, those fields are looked at only when you specify
+       SCM_PORT_WITH_POSITION.  They are needed to fully support srfi-192-style
+       port positioning.   See the Port Positioning Interface below. */
+    ScmObj (*GetPos)(ScmPort *p);
+    void   (*SetPos)(ScmPort *p, ScmObj pos);
+    u_long flags;               /* reserved */
 } ScmPortVTable;
 
-/* The main port structure.
- * Regardless of the port type, the port structure caches at most
- * one character, in order to realize `peek-char' (Scheme) or `Ungetc' (C)
- * operation.   'scratch', 'scrcnt', and 'ungotten' fields are used for
- * that purpose, and outside routine shouldn't touch these fields.
- * See portapi.c for the detailed semantics.
+/* Input string port */
+typedef struct ScmPortInputStringRec {
+    const char *start;
+    const char *current;
+    const char *end;
+} ScmPortInputString;
+
+/* Note on Port Positioning Interface
+
+   Gauche historically supported lseek(2) style interface, which can handle
+   both getting and setting the position.  R6RS introduced
+   port-position and set-port-position! interface, which then adopted as
+   SRFI-192.  It requires different strategy than lseek interface.
+
+   1) SRFI-192 allows port position can be gotten but not be set, and that
+      needs to be queried by port-has-set-port-position!?.  With lseek API,
+      port position is either totally unavaiable or can be gotten/set.
+      (seek callback may raise an error when setting is not supported, but
+      it can't be queried.)
+
+   2) SRFI-192 allows arbitrary Scheme object as the position, while lseek
+      API assumes integer byte offset.  Indeed, for the textual port, meaning
+      of byte offset is vague.
+
+   3) Lseek allows position to be set relative to the current position or
+      to the end of the file.  SRFI-192 set-port-position! is always the
+      absolute position.
+
+  To support both interface, we have these hooks in SCM_PORT_FILE and
+  SCM_PORT_PROC type of ports (the field names differ for the historical
+  reasons):
+
+     off_t SEEK(ScmPort*, off_t, int)    ;; seeker, Seek
+     ScmObj GETPOS(ScmPort*)             ;; getpos, GetPos
+     void SETPOS(ScmPort*, ScmObj)       ;; setpos, SetPos
+
+  GETPOS/SETPOS takes precedence to SEEK.
+
+  Port is port-has-port-position? if it has either GETPOS or SEEK handler,
+  and port-has-set-port-position!? if it has either SETPOS or SEEK handler.
+
+  For SCM_PORT_BUF, the position returned by GETPOS and that SETPOS takes
+  must be an exact Scheme integer, representing the integer byte offset.
+  Since the input is buffered, we need to calculate the actual position
+  in the buffer, from the source position returned by GETPOS.
+
+  For SCM_PORT_PROC, we treat the position as an opaque object.  Note that
+  input port has a small buffer for peeked byte/character, that makes
+  things complicated.  Since we can't adjust the opaque position, we instead
+  call GETPOS before peek and remember the position.
  */
 
+/*
+ * We don't expose much about ScmPort struct to the user.  A bunch of
+ * flags are public, but most of the stuff are hidden.  See priv/portP.h
+ * for the 'real' definition.
+ *
+ * ScmPort should never be allocated directly, or statically.  Port
+ * constructors properly initializes hidden fields.
+ */
+
+#define SCM_PORT_HEADER                                                 \
+    SCM_INSTANCE_HEADER;                                                \
+    u_int direction : 2;        /* SCM_PORT_INPUT or SCM_PORT_OUTPUT.   \
+                                   There may be I/O port in future. */  \
+    u_int type      : 2;        /* SCM_PORT_{FILE|ISTR|OSTR|PROC} */    \
+    u_int scrcnt    : 3;        /* # of bytes in the scratch buffer */  \
+                                                                        \
+    u_int ownerp    : 1;        /* TRUE if this port owns underlying    \
+                                   file pointer */                      \
+    u_int closed    : 1;        /* TRUE if this port is closed */       \
+    u_int error     : 1;        /* Error has been occurred */           \
+                                                                        \
+    u_int flags     : 5         /* see ScmPortFlags below */
+
 struct ScmPortRec {
-    SCM_INSTANCE_HEADER;
-    u_int direction : 2;        /* SCM_PORT_INPUT or SCM_PORT_OUTPUT.
-                                   There may be I/O port in future. */
-    u_int type      : 2;        /* SCM_PORT_{FILE|ISTR|OSTR|PROC} */
-    u_int scrcnt    : 3;        /* # of bytes in the scratch buffer */
+    SCM_PORT_HEADER;
+    /* See portP.h for the real definition.  The magic number is chosen
+       so that ScmPort has enough size to contain the real struct.
 
-    u_int ownerp    : 1;        /* TRUE if this port owns underlying
-                                   file pointer */
-    u_int closed    : 1;        /* TRUE if this port is closed */
-    u_int error     : 1;        /* Error has been occurred */
-
-    u_int flags     : 5;        /* see ScmPortFlags below */
-
-    char scratch[SCM_CHAR_MAX_BYTES]; /* incomplete buffer */
-
-    ScmChar ungotten;           /* ungotten character.
-                                   SCM_CHAR_INVALID if empty. */
-    ScmObj name;                /* port's name.  Can be any Scheme object. */
-
-    ScmInternalFastlock lock;   /* for port mutex */
-    ScmVM *lockOwner;           /* for port mutex; owner of the lock */
-    int lockCount;              /* for port mutex; # of recursive locks */
-
-    ScmWriteState *writeState;  /* used internally */
-
-    /* Input counters.  these doesn't take account of ungetting and
-       seeking: Ungetting doesn't affect those counters (you can think
-       that ungetting are handled above the counting layer).
-       Seeking invalidates counters; if you seek, the values of the counters
-       become bogus.
-       We don't have character counter, since it is difficult to track
-       (read-line uses byte read; see Scm_ReadLine in portapi.c).
+       We tested to use a pointer to the real struct, but I/O intensive
+       benchmark showed 2-3% reduction of speed.  So we avoid indirection.
      */
-    u_long line;                /* line counter */
-    u_long bytes;               /* byte counter */
-
-    /* The source or the sink of the port. */
-    union {
-        ScmPortBuffer buf;      /* buffered port */
-        struct {
-            const char *start;
-            const char *current;
-            const char *end;
-        } istr;                 /* input string port */
-        ScmDString ostr;        /* output string port */
-        ScmPortVTable vt;       /* virtual port */
-    } src;
-
-    /* Port attibutes.
-     * NB: Before we release 0.9.4, we might merge this into port->data and/or
-     * port->name.
-     */
-    ScmObj attrs;               /* port attibutes.  use Scm_PortAttr* API to
-                                   access. */
+    char opaque[29*sizeof(ScmWord) + 2*sizeof(ScmInternalFastlock)];
 };
 
-/* Port direction.  Bidirectional port is not supported yet.
+/* Port direction.
+   Bidirectional port have both SCM_PORT_INPUT and SCM_PORT_OUTPUT bits set.
    SCM_PORT_OUTPUT_TRANSIENT is only used for constructor to indicate
    that the output buffer doesn't need to be flushed when the process
    exits.  It is same as SCM_PORT_OUTPUT for non-buffered ports.
@@ -227,6 +254,13 @@ enum ScmFdReadyResult {
     SCM_FD_UNKNOWN
 };
 
+/* The flag bits passed to 'flags' argument of various port constructors */
+enum ScmPortConstructorFlag {
+    SCM_PORT_OWNER = (1L<<0),        /* Set the ownerp flag of the port */
+    SCM_PORT_WITH_POSITION = (1L<<1) /* The additional getpos/setpos fields
+                                        are looked at. */
+};
+
 /* Other flags used internally */
 /* NB: The first two flags only matter when port->recursiveContext is set,
    and they're transient by nature.  See write.c for the details. */
@@ -243,20 +277,6 @@ enum ScmPortFlags {
                                     registered for flushing. */
 };
 
-#if 0 /* not implemented */
-/* Incomplete character handling policy.
-   When Scm_Getc encounters a byte sequence that doesn't consist a valid
-   multibyte character, it may take one of the following actions,
-   according to the port's icpolicy field. */
-enum ScmPortICPolicy {
-    SCM_PORT_IC_ERROR,          /* signal an error */
-    SCM_PORT_IC_IGNORE,         /* ignore bytes until Getc finds a
-                                   valid multibyte character */
-    SCM_PORT_IC_REPLACE,        /* replace invalid byte to a designated
-                                   character. */
-};
-#endif
-
 /*================================================================
  * Generic operations
  */
@@ -268,7 +288,6 @@ enum ScmPortICPolicy {
 #define SCM_PORT_TYPE(obj)      (SCM_PORT(obj)->type)
 #define SCM_PORT_DIR(obj)       (SCM_PORT(obj)->direction)
 #define SCM_PORT_FLAGS(obj)     (SCM_PORT(obj)->flags)
-#define SCM_PORT_ICPOLICY(obj)  (SCM_PORT(obj)->icpolicy)
 
 #define SCM_PORT_CASE_FOLDING(obj) (SCM_PORT_FLAGS(obj)&SCM_PORT_CASE_FOLD)
 
@@ -276,22 +295,25 @@ enum ScmPortICPolicy {
 #define SCM_PORT_OWNER_P(obj)   (SCM_PORT(obj)->ownerp)
 #define SCM_PORT_ERROR_OCCURRED_P(obj) (SCM_PORT(obj)->error)
 
-#define SCM_PORT_BUFFER_DATA(obj)  (SCM_PORT(obj)->src.buf.data)
-#define SCM_PORT_VIRTUAL_DATA(obj) (SCM_PORT(obj)->src.vt.data)
-
 #define SCM_IPORTP(obj)  (SCM_PORTP(obj)&&(SCM_PORT_DIR(obj)&SCM_PORT_INPUT))
 #define SCM_OPORTP(obj)  (SCM_PORTP(obj)&&(SCM_PORT_DIR(obj)&SCM_PORT_OUTPUT))
 
+/* The following macros are for the backward compatibility; their use is
+   deprecated.  (We used to have these macros directly access ScmPort
+   structure to eliminate function call overhead.  It turned out
+   it matters little.   Use function APIs.) */
 #define SCM_PUTB(b, p)     Scm_Putb(b, SCM_PORT(p))
 #define SCM_PUTC(c, p)     Scm_Putc(c, SCM_PORT(p))
 #define SCM_PUTZ(s, l, p)  Scm_Putz(s, l, SCM_PORT(p))
 #define SCM_PUTS(s, p)     Scm_Puts(SCM_STRING(s), SCM_PORT(p))
 #define SCM_FLUSH(p)       Scm_Flush(SCM_PORT(p))
 #define SCM_PUTNL(p)       SCM_PUTC('\n', p)
-
 #define SCM_UNGETC(c, port) Scm_Ungetc(c, SCM_PORT(port))
 #define SCM_GETB(b, p)     (b = Scm_Getb(SCM_PORT(p)))
 #define SCM_GETC(c, p)     (c = Scm_Getc(SCM_PORT(p)))
+#define SCM_PORT_BUFFER_ROOM(p)  Scm_PortBufferRoom(p)
+#define SCM_PORT_BUFFER_AVAIL(p) Scm_PortBufferAvail(p)
+
 
 SCM_CLASS_DECL(Scm_PortClass);
 #define SCM_CLASS_PORT                (&Scm_PortClass)
@@ -319,10 +341,18 @@ SCM_EXTERN void   Scm_SetPortCaseFolding(ScmPort *port, int flag);
 SCM_EXTERN ScmObj Scm_GetPortReaderLexicalMode(ScmPort *port);
 SCM_EXTERN void   Scm_SetPortReaderLexicalMode(ScmPort *port, ScmObj obj);
 
+SCM_EXTERN void   Scm_SetPortErrorOccurred(ScmPort *port, int val);
+
+SCM_EXTERN int    Scm_PortPositionable(ScmPort *port, int setp);
+
 SCM_EXTERN void   Scm_FlushAllPorts(int exitting);
 
 SCM_EXTERN ScmObj Scm_PortName(ScmPort *port);
-SCM_EXTERN int    Scm_PortLine(ScmPort *port);
+SCM_EXTERN ScmSize Scm_PortLine(ScmPort *port);
+SCM_EXTERN ScmSize Scm_PortBytes(ScmPort *port);
+SCM_EXTERN ScmObj Scm_GetPortPosition(ScmPort *port);
+SCM_EXTERN ScmObj Scm_GetPortPositionUnsafe(ScmPort *port);
+SCM_EXTERN ScmObj Scm_SetPortPosition(ScmPort *port, ScmObj pos);
 SCM_EXTERN ScmObj Scm_PortSeek(ScmPort *port, ScmObj off, int whence);
 SCM_EXTERN ScmObj Scm_PortSeekUnsafe(ScmPort *port, ScmObj off, int whence);
 SCM_EXTERN int    Scm_PortFileNo(ScmPort *port);
@@ -333,6 +363,17 @@ SCM_EXTERN int    Scm_ByteReadyUnsafe(ScmPort *port);
 SCM_EXTERN int    Scm_CharReady(ScmPort *port);
 SCM_EXTERN int    Scm_CharReadyUnsafe(ScmPort *port);
 
+SCM_EXTERN ScmPortBuffer      *Scm_PortBufferStruct(ScmPort *port);
+SCM_EXTERN ScmPortInputString *Scm_PortInputStringStruct(ScmPort *port);
+SCM_EXTERN ScmDString         *Scm_PortOutputDString(ScmPort *port);
+SCM_EXTERN ScmPortVTable      *Scm_PortVTableStruct(ScmPort *port);
+
+SCM_EXTERN ScmSize  Scm_PortBufferRoom(ScmPort *port);
+SCM_EXTERN ScmSize  Scm_PortBufferAvail(ScmPort *port);
+
+SCM_EXTERN ScmWriteState *Scm_PortWriteState(ScmPort *);
+SCM_EXTERN void           Scm_PortWriteStateSet(ScmPort *, ScmWriteState*);
+
 SCM_EXTERN void   Scm_ClosePort(ScmPort *port);
 
 SCM_EXTERN ScmObj Scm_VMWithPortLocking(ScmPort *port,
@@ -340,16 +381,9 @@ SCM_EXTERN ScmObj Scm_VMWithPortLocking(ScmPort *port,
 
 SCM_EXTERN ScmObj Scm_PortAttrGet(ScmPort *port, ScmObj key,
                                   ScmObj fallback);
-SCM_EXTERN ScmObj Scm_PortAttrGetUnsafe(ScmPort *port, ScmObj key,
-                                        ScmObj fallback);
 SCM_EXTERN ScmObj Scm_PortAttrSet(ScmPort *port, ScmObj key, ScmObj val);
-SCM_EXTERN ScmObj Scm_PortAttrSetUnsafe(ScmPort *port, ScmObj key, ScmObj val);
-SCM_EXTERN ScmObj Scm_PortAttrCreate(ScmPort *port, ScmObj key, ScmObj get, ScmObj set);
-SCM_EXTERN ScmObj Scm_PortAttrCreateUnsafe(ScmPort *port, ScmObj key, ScmObj get, ScmObj set);
 SCM_EXTERN ScmObj Scm_PortAttrDelete(ScmPort *port, ScmObj key);
-SCM_EXTERN ScmObj Scm_PortAttrDeleteUnsafe(ScmPort *port, ScmObj key);
 SCM_EXTERN ScmObj Scm_PortAttrs(ScmPort *port);
-SCM_EXTERN ScmObj Scm_PortAttrsUnsafe(ScmPort *port);
 
 SCM_EXTERN void   Scm_Putb(ScmByte b, ScmPort *port);
 SCM_EXTERN void   Scm_Putc(ScmChar c, ScmPort *port);
@@ -413,6 +447,17 @@ SCM_EXTERN ScmObj Scm_SetCurrentErrorPort(ScmPort *port);
  * String ports
  */
 
+enum {
+    SCM_PORT_STRING_PRIVATE = 1 /* for flags arg */
+};
+
+SCM_EXTERN ScmObj Scm_MakeInputStringPortFull(ScmString *str,
+                                              ScmObj name,
+                                              u_long flags);
+SCM_EXTERN ScmObj Scm_MakeOutputStringPortFull(ScmObj name,
+                                               u_long flags);
+
+/* these two are deprecated */
 SCM_EXTERN ScmObj Scm_MakeInputStringPort(ScmString *str, int privatep);
 SCM_EXTERN ScmObj Scm_MakeOutputStringPort(int privatep);
 
@@ -424,13 +469,28 @@ SCM_EXTERN ScmObj Scm_GetRemainingInputString(ScmPort *port, int flags);
  * Other type of ports
  */
 
+SCM_EXTERN ScmObj Scm_MakeVirtualPortFull(ScmClass *klass,
+                                          ScmObj name, int direction,
+                                          const ScmPortVTable *vtable,
+                                          u_long flags);
+
+/* deprecated */
 SCM_EXTERN ScmObj Scm_MakeVirtualPort(ScmClass *klass,
                                       int direction,
                                       const ScmPortVTable *vtable);
+
+
+SCM_EXTERN ScmObj Scm_MakeBufferedPortFull(ScmClass *klass,
+                                           ScmObj name, int direction,
+                                           ScmPortBuffer *bufrec,
+                                           u_long flags);
+
+/* deprecated */
 SCM_EXTERN ScmObj Scm_MakeBufferedPort(ScmClass *klass,
                                        ScmObj name, int direction,
                                        int ownerp,
                                        ScmPortBuffer *bufrec);
+
 SCM_EXTERN ScmObj Scm_MakePortWithFd(ScmObj name,
                                      int direction,
                                      int fd,
@@ -439,4 +499,3 @@ SCM_EXTERN ScmObj Scm_MakePortWithFd(ScmObj name,
 SCM_EXTERN ScmObj Scm_MakeCodingAwarePort(ScmPort *iport);
 
 #endif /*GAUCHE_PORT_H*/
-
