@@ -1,7 +1,7 @@
 /*
  * gauche/vector.h - Vector API
  *
- *   Copyright (c) 2000-2018  Shiro Kawai  <shiro@acm.org>
+ *   Copyright (c) 2000-2022  Shiro Kawai  <shiro@acm.org>
  *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -40,18 +40,43 @@
 
 struct ScmVectorRec {
     SCM_HEADER;
+#if GAUCHE_API_VERSION >= 98
+    ScmWord size_flags;
+#else  /* GAUCHE_API_VERSION < 98 */
     ScmWord size;
+#endif /* GAUCHE_API_VERSION < 98 */
     ScmObj elements[1];
 };
 
+SCM_CLASS_DECL(Scm_VectorClass);
+#define SCM_CLASS_VECTOR     (&Scm_VectorClass)
 #define SCM_VECTOR(obj)          ((ScmVector*)(obj))
 #define SCM_VECTORP(obj)         SCM_XTYPEP(obj, SCM_CLASS_VECTOR)
-#define SCM_VECTOR_SIZE(obj)     (SCM_VECTOR(obj)->size)
 #define SCM_VECTOR_ELEMENTS(obj) (SCM_VECTOR(obj)->elements)
 #define SCM_VECTOR_ELEMENT(obj, i)   (SCM_VECTOR(obj)->elements[i])
 
-SCM_CLASS_DECL(Scm_VectorClass);
-#define SCM_CLASS_VECTOR     (&Scm_VectorClass)
+/* SCM_VECTOR_SIZE_SLOT_INITIALIZER is used in cgen-generated code  */
+
+#if GAUCHE_API_VERSION >= 98
+#define SCM_VECTOR_SIZE(obj)        (SCM_VECTOR(obj)->size_flags >> 1)
+#define SCM_VECTOR_IMMUTABLE_P(obj) (SCM_VECTOR(obj)->size_flags & 1)
+#define SCM_VECTOR_IMMUTABLE_SET(obj, flag)     \
+    ((flag)                                     \
+     ? (SCM_UVECTOR(obj)->size_flags |= 1)      \
+     : (SCM_UVECTOR(obj)->size_flags &= ~1))
+#define SCM_VECTOR_SIZE_SLOT_INITIALIZER(len, imm) \
+    SCM_OBJ(SCM_WORD(((len)<<1)|((imm)?1:0)))
+#else  /* GAUCHE_API_VERSION < 98 */
+#define SCM_VECTOR_SIZE(obj)        (SCM_VECTOR(obj)->size)
+#define SCM_VECTOR_IMMUTABLE_P(obj) (!SCM_VECTORP(obj)) /* always FALSE, but need to use obj to avoid unused variable warning */
+#define SCM_VECTOR_IMMUTABLE_SET(obj, flag)  /*empty*/
+#define SCM_VECTOR_SIZE_SLOT_INITIALIZER(len, imm)  SCM_OBJ(len)
+#endif /* GAUCHE_API_VERSION < 98 */
+
+#define SCM_VECTOR_CHECK_MUTABLE(obj)           \
+  do { if (SCM_VECTOR_IMMUTABLE_P(obj)) {       \
+    Scm_Error("vector is immutable: %S", obj);  \
+  }} while (0)
 
 SCM_EXTERN ScmObj Scm_MakeVector(ScmSmallInt size, ScmObj fill);
 SCM_EXTERN ScmObj Scm_VectorRef(ScmVector *vec, ScmSmallInt i, ScmObj fallback);
@@ -123,6 +148,11 @@ typedef enum {
     SCM_UVECTOR_F16,
     SCM_UVECTOR_F32,
     SCM_UVECTOR_F64,
+    SCM_UVECTOR_RESERVED1,      /* reserved for f128*/
+    SCM_UVECTOR_C32,
+    SCM_UVECTOR_C64,
+    SCM_UVECTOR_C128,
+    SCM_UVECTOR_RESERVED2,      /* reserved for c256 */
     SCM_UVECTOR_INVALID = -1
 } ScmUVectorType;
 
@@ -232,26 +262,26 @@ SCM_CLASS_DECL(Scm_S64VectorClass);
 #define SCM_S64VECTOR(obj)          SCM_UVECTOR(obj)
 #define SCM_S64VECTORP(obj)         SCM_XTYPEP(obj, SCM_CLASS_S64VECTOR)
 #define SCM_S64VECTOR_SIZE(obj)     SCM_UVECTOR_SIZE(obj)
-#define SCM_S64VECTOR_ELEMENTS(obj) ((ScmInt64*)SCM_UVECTOR_ELEMENTS(obj))
+#define SCM_S64VECTOR_ELEMENTS(obj) ((int64_t*)SCM_UVECTOR_ELEMENTS(obj))
 #define SCM_S64VECTOR_ELEMENT(obj,k) SCM_S64VECTOR_ELEMENTS(obj)[k]
-SCM_EXTERN ScmObj Scm_MakeS64Vector(ScmSmallInt size, ScmInt64 fill);
+SCM_EXTERN ScmObj Scm_MakeS64Vector(ScmSmallInt size, int64_t fill);
 SCM_EXTERN ScmObj Scm_MakeS64VectorFromArray(ScmSmallInt size,
-                                             const ScmInt64 array[]);
+                                             const int64_t array[]);
 SCM_EXTERN ScmObj Scm_MakeS64VectorFromArrayShared(ScmSmallInt size,
-                                                   ScmInt64 array[]);
+                                                   int64_t array[]);
 
 SCM_CLASS_DECL(Scm_U64VectorClass);
 #define SCM_CLASS_U64VECTOR         (&Scm_U64VectorClass)
 #define SCM_U64VECTOR(obj)          SCM_UVECTOR(obj)
 #define SCM_U64VECTORP(obj)         SCM_XTYPEP(obj, SCM_CLASS_U64VECTOR)
 #define SCM_U64VECTOR_SIZE(obj)     SCM_UVECTOR_SIZE(obj)
-#define SCM_U64VECTOR_ELEMENTS(obj) ((ScmUInt64*)SCM_UVECTOR_ELEMENTS(obj))
+#define SCM_U64VECTOR_ELEMENTS(obj) ((uint64_t*)SCM_UVECTOR_ELEMENTS(obj))
 #define SCM_U64VECTOR_ELEMENT(obj,k) SCM_U64VECTOR_ELEMENTS(obj)[k]
-SCM_EXTERN ScmObj Scm_MakeU64Vector(ScmSmallInt size, ScmUInt64 fill);
+SCM_EXTERN ScmObj Scm_MakeU64Vector(ScmSmallInt size, uint64_t fill);
 SCM_EXTERN ScmObj Scm_MakeU64VectorFromArray(ScmSmallInt size,
-                                             const ScmUInt64 array[]);
+                                             const uint64_t array[]);
 SCM_EXTERN ScmObj Scm_MakeU64VectorFromArrayShared(ScmSmallInt size,
-                                                   ScmUInt64 array[]);
+                                                   uint64_t array[]);
 
 SCM_CLASS_DECL(Scm_F16VectorClass);
 #define SCM_CLASS_F16VECTOR         (&Scm_F16VectorClass)
@@ -292,8 +322,48 @@ SCM_EXTERN ScmObj Scm_MakeF64VectorFromArray(ScmSmallInt size,
 SCM_EXTERN ScmObj Scm_MakeF64VectorFromArrayShared(ScmSmallInt size,
                                                    double array[]);
 
+SCM_CLASS_DECL(Scm_C32VectorClass);
+#define SCM_CLASS_C32VECTOR         (&Scm_C32VectorClass)
+#define SCM_C32VECTOR(obj)          SCM_UVECTOR(obj)
+#define SCM_C32VECTORP(obj)         SCM_XTYPEP(obj, SCM_CLASS_C32VECTOR)
+#define SCM_C32VECTOR_SIZE(obj)     SCM_UVECTOR_SIZE(obj)
+#define SCM_C32VECTOR_ELEMENTS(obj) ((ScmHalfComplex*)SCM_UVECTOR_ELEMENTS(obj))
+#define SCM_C32VECTOR_ELEMENT(obj,k) SCM_C32VECTOR_ELEMENTS(obj)[k]
+SCM_EXTERN ScmObj Scm_MakeC32Vector(ScmSmallInt size, ScmHalfComplex fill);
+SCM_EXTERN ScmObj Scm_MakeC32VectorFromArray(ScmSmallInt size,
+                                             const ScmHalfComplex array[]);
+SCM_EXTERN ScmObj Scm_MakeC32VectorFromArrayShared(ScmSmallInt size,
+                                                   ScmHalfComplex array[]);
 
-/* For the backward compatibility */
+SCM_CLASS_DECL(Scm_C64VectorClass);
+#define SCM_CLASS_C64VECTOR         (&Scm_C64VectorClass)
+#define SCM_C64VECTOR(obj)          SCM_UVECTOR(obj)
+#define SCM_C64VECTORP(obj)         SCM_XTYPEP(obj, SCM_CLASS_C64VECTOR)
+#define SCM_C64VECTOR_SIZE(obj)     SCM_UVECTOR_SIZE(obj)
+#define SCM_C64VECTOR_ELEMENTS(obj) ((ScmFloatComplex*)SCM_UVECTOR_ELEMENTS(obj))
+#define SCM_C64VECTOR_ELEMENT(obj,k) SCM_C64VECTOR_ELEMENTS(obj)[k]
+SCM_EXTERN ScmObj Scm_MakeC64Vector(ScmSmallInt size, ScmFloatComplex fill);
+SCM_EXTERN ScmObj Scm_MakeC64VectorFromArray(ScmSmallInt size,
+                                             const ScmFloatComplex array[]);
+SCM_EXTERN ScmObj Scm_MakeC64VectorFromArrayShared(ScmSmallInt size,
+                                                   ScmFloatComplex array[]);
+
+SCM_CLASS_DECL(Scm_C128VectorClass);
+#define SCM_CLASS_C128VECTOR         (&Scm_C128VectorClass)
+#define SCM_C128VECTOR(obj)          SCM_UVECTOR(obj)
+#define SCM_C128VECTORP(obj)         SCM_XTYPEP(obj, SCM_CLASS_C128VECTOR)
+#define SCM_C128VECTOR_SIZE(obj)     SCM_UVECTOR_SIZE(obj)
+#define SCM_C128VECTOR_ELEMENTS(obj) ((ScmDoubleComplex*)SCM_UVECTOR_ELEMENTS(obj))
+#define SCM_C128VECTOR_ELEMENT(obj,k) SCM_C128VECTOR_ELEMENTS(obj)[k]
+SCM_EXTERN ScmObj Scm_MakeC128Vector(ScmSmallInt size, ScmDoubleComplex fill);
+SCM_EXTERN ScmObj Scm_MakeC128VectorFromArray(ScmSmallInt size,
+                                             const ScmDoubleComplex array[]);
+SCM_EXTERN ScmObj Scm_MakeC128VectorFromArrayShared(ScmSmallInt size,
+                                                   ScmDoubleComplex array[]);
+
+
+/* For the backward compatibility.  Use ScmUVector for the new code
+   (that's why we don't define ScmC**Vector, for they are new additions. */
 typedef ScmUVector ScmS8Vector;
 typedef ScmUVector ScmU8Vector;
 typedef ScmUVector ScmS16Vector;
@@ -306,12 +376,64 @@ typedef ScmUVector ScmF16Vector;
 typedef ScmUVector ScmF32Vector;
 typedef ScmUVector ScmF64Vector;
 
+/*
+ * View memory as an uvector
+ */
+ScmObj Scm_MakeViewUVector(ScmMemoryRegion *mem, ScmClass *klass,
+                           ScmSmallInt len, ScmSmallInt offset,
+                           int immutable);
 
-/* 
+/*
  * String/uvector common utility
  */
 
 /* Retrieves underlying byte array */
 const uint8_t *Scm_GetBytes(ScmObj str_or_uvec, ScmSize *size);
 
+/*
+ * Bitvectors (srfi-178)
+ *
+ *   We could view a bitvector as u1vector, but the operations on bitvectors
+ *   are sufficiently different from uniform vectors, so we made them
+ *   disjoint.
+ */
+
+typedef struct ScmBitvectorRec {
+    SCM_HEADER;
+    ScmWord size_flags;         /* (len<<1)|immutable */
+    ScmBits *bits;
+} ScmBitvector;
+
+SCM_CLASS_DECL(Scm_BitvectorClass);
+#define SCM_CLASS_BITVECTOR         (&Scm_BitvectorClass)
+#define SCM_BITVECTOR(obj)          ((ScmBitvector*)(obj))
+#define SCM_BITVECTORP(obj)         Scm_TypeP(obj, SCM_CLASS_BITVECTOR)
+#define SCM_BITVECTOR_BITS(obj)     (SCM_BITVECTOR(obj)->bits)
+
+#define SCM_BITVECTOR_SIZE(obj)     (SCM_BITVECTOR(obj)->size_flags >> 1)
+#define SCM_BITVECTOR_IMMUTABLE_P(obj) (SCM_BITVECTOR(obj)->size_flags & 1)
+#define SCM_BITVECTOR_IMMUTABLE_SET(obj, flag)    \
+    ((flag)                                     \
+     ? (SCM_BITVECTOR(obj)->size_flags |= 1)      \
+     : (SCM_BITVECTOR(obj)->size_flags &= ~1))
+
+#define SCM_BITVECTOR_CHECK_MUTABLE(obj)                 \
+  do { if (SCM_BITVECTOR_IMMUTABLE_P(obj)) {             \
+    Scm_Error("bitvector is immutable: %S", obj); \
+  }} while (0)
+
+SCM_EXTERN int    Scm_Bit2Int(ScmObj bit);
+SCM_EXTERN ScmObj Scm_Bit2Bool(ScmObj bit);
+SCM_EXTERN ScmObj Scm_MakeBitvector(ScmSmallInt size, ScmObj init);
+SCM_EXTERN ScmObj Scm_ListToBitvector(ScmObj lis);
+SCM_EXTERN ScmObj Scm_BitvectorCopy(ScmBitvector *v,
+                                    ScmSmallInt start,
+                                    ScmSmallInt end);
+SCM_EXTERN ScmObj Scm_BitvectorCopyX(ScmBitvector *dest, ScmSmallInt dstart,
+                                     ScmBitvector *src,
+                                     ScmSmallInt sstart, ScmSmallInt send);
+
+SCM_EXTERN ScmObj Scm_StringToBitvector(ScmString *s, int prefix);
+SCM_EXTERN ScmObj Scm_BitvectorToString(ScmBitvector *v, int prefix);
+
 #endif /*GAUCHE_VECTOR_H*/
