@@ -1,7 +1,7 @@
 /*
  * vm.h - Virtual machine
  *
- *   Copyright (c) 2000-2021  Shiro Kawai  <shiro@acm.org>
+ *   Copyright (c) 2000-2022  Shiro Kawai  <shiro@acm.org>
  *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -225,6 +225,11 @@ SCM_EXTERN int    Scm_IdentifierBindingEqv(ScmIdentifier *id,
 SCM_EXTERN ScmIdentifier *Scm_OutermostIdentifier(ScmIdentifier *id);
 SCM_EXTERN ScmSymbol     *Scm_UnwrapIdentifier(ScmIdentifier *id);
 SCM_EXTERN ScmGloc       *Scm_IdentifierGlobalBinding(ScmIdentifier *id);
+SCM_EXTERN ScmObj         Scm_IdentifierGlobalRef(ScmIdentifier *id,
+                                                  ScmGloc **pgloc);
+SCM_EXTERN void           Scm_IdentifierGlobalSet(ScmIdentifier *id,
+                                                  ScmObj val,
+                                                  ScmGloc **pgloc);
 
 /*
  * Escape handling
