@@ -1,7 +1,7 @@
 ;;;
 ;;;  data.heap - Heaps
 ;;;
-;;;   Copyright (c) 2014-2019  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2014-2022  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -34,7 +34,7 @@
 (define-module data.heap
   (use gauche.sequence)
   (use gauche.uvector)
-  (use srfi-1)
+  (use scheme.list)
   (export <binary-heap>
           make-binary-heap build-binary-heap
           binary-heap-comparator binary-heap-key-procedure
@@ -66,7 +66,7 @@
 ;; Min-max heaps and generalized priority queues, CACM 29(10) pp.996-1000
 ;; Oct. 1986, http://dx.doi.org/10.1145/6617.6621
 
-;; Binary heap is stored in an array, S[].  
+;; Binary heap is stored in an array, S[].
 ;;
 ;; - S[1] is the root.
 ;; - S[i]'s child is S[2i] and S[2i+1]
@@ -131,7 +131,7 @@
           :<: <:
           :>: >:)))
 
-;; heapify 
+;; heapify
 (define (build-binary-heap storage :key (num-entries #f)
                                         (comparator default-comparator)
                                         (key identity))
@@ -150,14 +150,14 @@
                      [(and (integer? num-entries) (>= num-entries 0))
                       (min num-entries xlen)]
                      [else
-                      (error "invalid num-entris value for build-binary-heap:"
+                      (error "invalid num-entries value for build-binary-heap:"
                              num-entries)])])
     (receive (<: >:)
         (ecase (comparator-flavor comparator)
           [(ordering) (values (comparator-ordering-predicate comparator)
                               (^[a b] (>? comparator a b)))]
           [(comparison) (values (^[a b] (<? comparator a b))
-                                (^[a b] (>? comparator a b)))])        
+                                (^[a b] (>? comparator a b)))])
       (bh-heapify! storage <: >: size)
       (make <binary-heap> :comparator comparator :storage storage :key key
             :<: <: :>: >: :next-leaf (+ size 1)
@@ -289,12 +289,19 @@
 
 ;; not exactly heap operations, but useful...
 
-(define (binary-heap-find hp pred)
-  (and-let* ([storage (~ hp'storage)]
-             [i (find-index (^i (pred (~ storage i)))
-                            (liota (binary-heap-num-entries hp)))])
-    ;; NB: This i is 0-base, so we don't need Ix.
-    (~ storage i)))
+;; Up to 0.9.10, we have (binary-heap-find <heap> <pred>) but that doesn't
+;; align with other *-find procedures.  In 0.9.11 we change it to
+;; (binary-heap-find <pred> <heap> :optional <failure>), while supporting
+;; old API as well.
+(define (binary-heap-find pred hp :optional (failure (^[] #f)))
+  (if (is-a? pred <binary-heap>)
+    (binary-heap-find hp pred failure) ; backward compatibility
+    (or (and-let* ([storage (~ hp'storage)]
+                   [i (find-index (^i (pred (~ storage i)))
+                                  (liota (binary-heap-num-entries hp)))])
+          ;; NB: This i is 0-base, so we don't need Ix.
+          (~ storage i))
+        (failure))))
 
 (define (binary-heap-remove! hp pred)
   (let1 storage (~ hp'storage)
@@ -390,7 +397,7 @@
     (bh-bubble-up storage <: >: i)))
 
 (define (bh-trickle-down storage <: >: index size)
-  
+
   (define-syntax getval (syntax-rules () [(getval i) (~ storage (Ix i))]))
   (define-syntax in-bound? (syntax-rules () [(in-bound? i) (< i size)]))
 
@@ -430,4 +437,3 @@
           (trickle-down-rec >< pick)))))
 
   (trickle-down-rec (if (min-node? index) <: >:) index))
-
