@@ -1,7 +1,7 @@
 ;;;
 ;;; gauche.cgen.cise - C in S expression
 ;;;
-;;;   Copyright (c) 2004-2018  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2004-2022  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -34,7 +34,6 @@
 (define-module gauche.cgen.cise
   (use srfi-13)
   (use gauche.sequence)
-  (use gauche.parameter)
   (use gauche.cgen.unit)
   (use gauche.cgen.literal)
   (use gauche.experimental.lamb)
@@ -372,7 +371,6 @@
     [(x . y)     form]   ; already stree
     ['|#reset-line| '|#reset-line|] ; special directive to reset line info
     [[? type-decl-initial?] (wrap-expr (cise-render-typed-var form "" env) env)]
-    [[? symbol?] (wrap-expr (cise-render-identifier form) env)]
     [[? identifier?] (wrap-expr (cise-render-identifier (unwrap-syntax form))
                                 env)]
     [[? string?] (wrap-expr (write-to-string form) env)]
@@ -453,8 +451,18 @@
 (define (expand-cfn form env)
   (define (gen-args args env)
     (let1 eenv (expr-env env)
-      ($ intersperse ","
-         $ map (^.[(var . type) (cise-render-typed-var type var eenv)]) args)))
+      (define (gen-arg-1 var type)
+        (if (eq? var '_)
+          `(,@(cise-render-typed-var type (gensym) eenv) " SCM_UNUSED")
+          (cise-render-typed-var type var eenv)))
+      (let loop ([args args])
+        (match args
+          [() '()]
+          [(('... . _)) '("...")]       ;varargs
+          [((var . type))
+           `(,@(gen-arg-1 var type))]
+          [((var . type) . rest)
+           `(,@(gen-arg-1 var type) "," ,@(loop rest))]))))
 
   (define (gen-qualifiers quals) ; we might support more qualifiers in future
     (intersperse " "
@@ -504,9 +512,9 @@
          [(declare-cfn)
           (unless (null? body)
             (errorf "declare-cfn ~s must not have a body" name))
-          (when (or (memq :static quals) (memq :inline quals))
-            (errorf "declare-cfn ~s cannot have qualifier(s)" name))
-          (record-static name '(:extern) args ret-type)
+          (if (and (not (memq :extern quals)) (not (memq :static quals)))
+            (record-static name `(:extern ,@quals) args ret-type)
+            (record-static name quals args ret-type))
           ;; no function implementation
           '()])]))
 
@@ -554,24 +562,25 @@
       [(':static . init-and-quals)
        (check-quals var type `(:static ,@quals) init-and-quals)]
       [((? keyword? z) . body)
-       (errorf "Invalid qualifier in define-cvar ~s: ~s" var z)]
+       (errorf "Invalid qualifier in `define-cvar ~s': ~s" var z)]
       [()
-       (case (car form)
+       (ecase (car form)
          [(define-cvar) (gen-cvar var type quals #f #f)]
          [(declare-cvar)
           (unless (null? quals)
-            (errorf "declare-cvar ~s cannot have qualifier(s)" var))
+            (errorf "`declare-cvar ~s' cannot have qualifier(s)" var))
           (gen-cvar var type '(:extern) #f #f)]
          [(define-ctype)
           (unless (null? quals)
-            (errorf "define-ctype ~s cannot have qualifier(s)" var))
+            (errorf "`define-ctype ~s' cannot have qualifier(s)" var))
           (gen-cvar var type '(:typedef) #f #f)])]
       [(init)
-       (if (eq? (car form) 'define-cvar)
-         (gen-cvar var type quals #t init)  
-         (errorf "declare-cvar ~s cannot have initializer" var))]
+       (ecase (car form)
+         [(define-cvar) (gen-cvar var type quals #t init)]
+         [(declare-cvar define-ctype)
+          (errorf "`~a ~s' cannot have an initializer" (car form) var)])]
       [else
-       (errorf "Invalid syntax in ~s ~s: ~s"
+       (errorf "Invalid syntax in `~s ~s': ~s"
                (car form) var init-and-quals)]))
 
   ;; We allow define-cvar only on toplevel, but declare-cvar and
@@ -586,7 +595,7 @@
     (receive (type init-and-quals)
         (match spec
           [()         (values 'ScmObj '())]
-          [('::)      (errorf "invalid variable decl in ~s: (~s ~s)" 
+          [('::)      (errorf "invalid variable decl in ~s: (~s ~s)"
                               (car form) var spec)]
           [(':: type) (values type '())]
           [(':: type . init-and-quals) (values type init-and-quals)]
@@ -1277,6 +1286,8 @@
 (define-cise-expr C:
   [(_ stuff) (list (x->string stuff))])
 
+;; DEPRECATED: cgen-stub-cise-ambient overrides 'return' macro.
+;; Use it instead.
 (define-cise-expr result
   [(_) (error "cise: result form needs at least one value'")]
   [(_ e) `(set! SCM_RESULT ,e)]
@@ -1370,7 +1381,7 @@
          ")"))]
     [(x)
      `(,(x->string x) " " ,(cise-render-identifier var))]
-    [(x xs ...) 
+    [(x xs ...)
      `(,(x->string x) " " ,(cise-render-typed-var xs var env))]
     [x
      `(,(x->string x) " " ,(cise-render-identifier var))]))
@@ -1378,7 +1389,7 @@
 (define (render-struct-or-union struct/union tag fields rest var env)
   `(,struct/union
     ,@(cond-list [tag `(" " ,tag)]
-                 [fields 
+                 [fields
                   `(" { "
                     ,@(map (^.[(member ':: type)
                                `(,(cise-render-typed-var type member env)
@@ -1393,8 +1404,23 @@
 (define (cise-render-identifier sym)
   (cgen-safe-name-friendly (x->string sym)))
 
-;; Allow var::type as (var :: type)
+;; canonicalize-vardecl
+;;
+;; Given list of C variable declarations, possibly typed, returns a list
+;; ((var :: type [init]) ...)
+;;
+;; Allows var::type as (var :: type)
 ;; and (var::type init) as (var :: type init)
+;;
+;; It is also used to expand the field list of struct and union.
+;;
+;; Example: (foo bar::int baz::(const char*) (boom::int 0))
+;;  => ((foo :: ScmObj) (bar :: int) (baz :: (const char*)) (boom :: int 0))
+;;
+;; Note: To avoid any type decl to be generated, give || as a type, e.g.
+;; "foo::||".  It is unusual, but needed if the variable name is actually
+;; a macro to be expanded into typed variable name.  SCM_HEADER is an example.
+;; We need a placeholder || to avoid the default type ScmObj is given.
 (define (canonicalize-vardecl vardecls)
   (define (expand-type elt seed)
     (cond
@@ -1404,9 +1430,10 @@
         [else (cons elt seed)])]
      [(symbol? elt)
       (rxmatch-case (symbol->string elt)
-        [#/^(.+)::$/ (_ v) `(,(string->symbol v) :: ,@seed)]
-        [#/^(.+)::(.+)$/ (_ v t)
+        [#/^([^:]+)::$/ (_ v) `(,(string->symbol v) :: ,@seed)]
+        [#/^([^:]+)::([^:]+)$/ (_ v t)
             `(,(string->symbol v) :: ,(string->symbol t) ,@seed)]
+        [#/^([^:]+):(.*)$/ (#f #f #f) (err elt)]
         [else (cons elt seed)])]
      [else (cons elt seed)]))
 
@@ -1444,4 +1471,3 @@
 ;;   This must come at the bottom of the module.
 
 (cise-ambient (cise-default-ambient))
-
