@@ -1,7 +1,7 @@
 ;;;
 ;;; gauche.cgen.literal - static literal data
 ;;;
-;;;   Copyright (c) 2004-2018  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2004-2022  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -32,10 +32,9 @@
 ;;;
 
 (define-module gauche.cgen.literal
-  (use srfi-1)
+  (use scheme.list)
   (use srfi-13)
   (use srfi-42)
-  (use gauche.parameter)
   (use gauche.sequence)
   (use gauche.cgen.unit)
   (use util.match)
@@ -52,6 +51,9 @@
   )
 (select-module gauche.cgen.literal)
 
+;; <module> literal needs to access this
+(autoload gauche.cgen.tmodule all-tmodules)
+
 ;;=============================================================
 ;; Static objects
 ;;
@@ -119,6 +121,11 @@
         (or dl
             (rlet1 new (make <cgen-static-data-list>
                          :category category :c-type c-type)
+              (when (eq? c-type 'ScmPair)
+                ;; We need a sentinel at the beginning of ScmPair array
+                ;; See ScmExtendedPair description in gauche/priv/pairP.h.
+                (push! (~ new'init-thunks) "   { SCM_NIL, SCM_NIL }")
+                (inc! (~ new'count)))
               (push! (~ unit'static-data-list) new))))))
 
   (let ([dl (ensure-static-data-list category c-type)]
@@ -163,8 +170,11 @@
               name)
       (dolist [dl dls]
         (for-each (cut print "#if "<>) (reverse (~ dl'cpp-conditions)))
-        (format #t "  ~a ~a[~a];\n" (~ dl'c-type) (~ dl'c-member-name)
-                (~ dl'count))
+        (format #t "  ~a ~a[~a]~a;\n"
+                (~ dl'c-type) (~ dl'c-member-name) (~ dl'count)
+                (if (eq? (~ dl'c-type) 'ScmPair)
+                  " SCM_ALIGN_PAIR"
+                  ""))
         (for-each (cut print "#endif /*"<>"*/") (~ dl'cpp-conditions)))
       (format #t "} ~a SCM_UNUSED = " name)))
 
@@ -356,29 +366,34 @@
 ;; literal value management -----------------------------------
 
 ;; We want to share the same literals.  The criteria of this 'same' is
-;; a bit complicated, and we can't use a hashtable for it (unless we have
-;; a hashtable with completely customizable hash fn and cmp fn, which Gauche
-;; doesn't have yet).  So we roll our own table, at least for the time being.
+;; a bit complicated, and we need a customized comparator for thta.
 ;;
 ;; NOTE: We don't share literals when they have different cpp-conditions,
 ;; even if they are the same otherwise.  Theoretically we can share if one
 ;; has cpp-condition and the other doesn't, but tracking those dependencies
 ;; is just a headache.  We just assume such sharing rarely occurs.
+;; This, the key of the hashtable consists of a pair of the literal value
+;; and its cpp condition.
 
-(define-constant .literal-hash-size. 32769)
+(define (%descriptive-type? obj)
+  (is-a? obj <descriptive-type>))
 
-(define (literal-value-hash literal)
-  (define mask #x0fffffff)
+(define (%deconstruct-type t)
+  ((with-module gauche.internal deconstruct-type) t))
+
+(define (literal-value-hash key)
   (define (rec val)
     (cond
-     [(pair? val) (logand (+ (rec (car val)) (rec (cdr val))) mask)]
-     [(vector? val) (fold (^(v r) (logand (+ (rec v) r) mask)) 0 val)]
-     [(string? val) (logand (string-hash val) mask)]
-     [(identifier? val) (logand (+ (rec (~ val'name))(rec (~ val'module))) mask)]
-     [else (eqv-hash val)]))
-  (modulo (rec literal) .literal-hash-size.))
+     [(wrapped-identifier? val)
+      (combine-hash-value (rec (~ val'name)) (rec (~ val'module)))]
+     [(%descriptive-type? val)
+      (fold combine-hash-value (rec (~ val'name))
+            (map rec (%deconstruct-type val)))]
+     [else (default-hash val)]))
+  ;; for hashing, we ignore cpp condition
+  (rec (car key)))
 
-(define (literal-value=? x y)
+(define (literal-value=? x-key y-key)
   (define (rec x y)
     (cond
      [(pair? x) (and (pair? y) (rec (car x) (car y)) (rec (cdr x) (cdr y)))]
@@ -388,37 +403,43 @@
              (and (= len (vector-length y))
                   (every?-ec (: i len)
                              (rec (vector-ref x i) (vector-ref y i))))))]
-     [(string? x) (and (string? y) (string=? x y))]
-     [(identifier? x)
-      (and (identifier? y)
+     [(or (uvector? x) (string? x) (char-set? x) (regexp? x)
+          (%descriptive-type? x))
+      (equal? x y)]
+     [(wrapped-identifier? x)
+      (and (wrapped-identifier? y)
            (eq? (~ x'name) (~ y'name))
            (eq? (~ x'module) (~ y'module)))]
+     ;; We can't use equal? as fallback, for the equality predicate needs
+     ;; to recurse to literal-value=?.
      [else (and (eq? (class-of x) (class-of y)) (eqv? x y))]))
-  (rec x y))
+  (and (equal? (cdr x-key) (cdr y-key)) ;cpp conditions
+       (rec (car x-key) (car y-key))))
 
 (define (ensure-literal-hash unit)
   (or (~ unit'literals)
-      (rlet1 hash (make-vector .literal-hash-size. '())
+      (rlet1 hash (make-hash-table (make-comparator (constantly #t)
+                                                    literal-value=?
+                                                    #f
+                                                    literal-value-hash))
         (set! (~ unit'literals) hash))))
 
+(define *uniq* (cons #f #f))
+
 (define (register-literal-value unit literal-obj)
-  (let ([lh   (ensure-literal-hash unit)]
-        [cppc (~ literal-obj'cpp-conditions)]
-        [h    (literal-value-hash (~ literal-obj'value))])
-    (or (and-let* ([entry (find (^e (and (equal? (caar e) cppc)
-                                         (literal-value=? (~ literal-obj'value)
-                                                          (cdar e))))
-                                (vector-ref lh h))])
-          (set-cdr! entry literal-obj))
-        (push! (vector-ref lh h) (acons cppc (~ literal-obj'value) literal-obj)))))
+  (let ([h (ensure-literal-hash unit)]
+        [key (cons (~ literal-obj'value)
+                   (~ literal-obj'cpp-conditions))])
+    (hash-table-update! h key
+                        (^v (if (eq? v *uniq*)
+                              literal-obj
+                              v))
+                        *uniq*)))
 
 (define (lookup-literal-value unit val)
-  (let ([lh (ensure-literal-hash unit)]
-        [cppc (cgen-cpp-conditions)])
-    (and-let* ([entry (find (^e (and (equal? (caar e) cppc)
-                                     (literal-value=? val (cdar e))))
-                            (vector-ref lh (literal-value-hash val)))])
-      (cdr entry))))
+  (let ([h (ensure-literal-hash unit)]
+        [key (cons val (cgen-cpp-conditions))])
+    (hash-table-get h key #f)))
 
 ;; primitive values -------------------------------------------
 
@@ -475,7 +496,7 @@
                'constant 'ScmString
                (format "  SCM_STRING_CONST_INITIALIZER(~a, ~a, ~a)"
                        (cgen-safe-string value)
-                       (string-size value) (string-length value)))
+                       (string-length value) (string-size value)))
       :value value))
   )
 
@@ -617,12 +638,14 @@
    (cdr :init-keyword :cdr))
   (make (value)
     (let* ([ca (cgen-literal (car value))]
-           [cd (cgen-literal (cdr value))]
-           [sobj (cgen-allocate-static-array
-                  'runtime 'ScmObj
-                  (list (get-literal-initializer ca)
-                        (get-literal-initializer cd)))])
-      (make <cgen-scheme-pair> :value value :car ca :cdr cd :c-name sobj)))
+           [cd (cgen-literal (cdr value))])
+      (define (init-thunk)
+        (format #t "   { ~a, ~a}"
+                (get-literal-initializer ca)
+                (get-literal-initializer cd)))
+      (make <cgen-scheme-pair> :value value :car ca :cdr cd
+            :c-name (cgen-allocate-static-datum 'runtime 'ScmPair
+                                                init-thunk))))
   (init (self)
     (let1 cname (cgen-cexpr self)
       (unless (cgen-literal-static? (~ self'car))
@@ -680,7 +703,9 @@
                   'runtime 'ScmObj
                   (list*
                    "SCM_OBJ(SCM_CLASS_STATIC_TAG(Scm_VectorClass)) /* <vector> */"
-                   (format "SCM_OBJ(~a)" (length literals))
+                   (format "SCM_VECTOR_SIZE_SLOT_INITIALIZER(~a, ~a)"
+                           (length literals)
+                           (if (vector-immutable? value) "TRUE" "FALSE"))
                    (map (^[lit] (if (cgen-literal-static? lit)
                                   (cgen-cexpr lit)
                                   "SCM_UNDEFINED"))
@@ -717,11 +742,10 @@
   [decl (self)
     (let* ([value (~ self'value)]
            [class (class-of value)])
-      (print (uvector-class->c-type-name class)" "(~ self'elements)"[] = {")
+      (print "static "(uvector-class->c-type-name class)" "(~ self'elements)"[] = {")
       (dotimes [i (uvector-length value)]
         ($ uvector-class-emit-elt class
-           $ (with-module gauche.internal %uvector-ref)
-             value (uvector-class->type-enum class) i))
+           $ (with-module gauche.internal uvector-ref) value i))
       (print  "};"))]
   [static (self) #f]
   )
@@ -744,16 +768,35 @@
     [(eq? class <u8vector>) "unsigned char"]
     [(eq? class <s16vector>) "short"]
     [(eq? class <u16vector>) "unsigned short"]
-    [(eq? class <s32vector>) "ScmInt32"]
-    [(eq? class <u32vector>) "ScmUInt32"]
-    [(eq? class <s64vector>) "ScmInt64"]
-    [(eq? class <u64vector>) "ScmUInt64"]
+    [(eq? class <s32vector>) "int32_t"]
+    [(eq? class <u32vector>) "uint32_t"]
+    [(eq? class <s64vector>) "int64_t"]
+    [(eq? class <u64vector>) "uint64_t"]
     [(eq? class <f16vector>) "ScmHalfFloat"]
     [(eq? class <f32vector>) "float"]
     [(eq? class <f64vector>) "double"]
+    [(eq? class <c32vector>) "ScmHalfComplex"]
+    [(eq? class <c64vector>) "ScmFloatComplex"]
+    [(eq? class <c128vector>) "ScmDoubleComplex"]
     ))
 
 (define (uvector-class-emit-elt class v)
+  (define (pr-half-float v)
+    (format #t "0x~4,'0x" ((with-module gauche.internal flonum->f16bits) v)))
+  (define (pr-float v)
+    (cond [(nan? v) (display "SCM_FLT_NAN")]
+          [(infinite? v) (display (if (> v 0)
+                                    "SCM_FLT_POSITIVE_INFINITY"
+                                    "SCM_FLT_NEGATIVE_INFINITY"))]
+          [(-zero? v) (display "-0.0f")]
+          [else (format #t "~sf" v)]))
+  (define (pr-double v)
+    (cond [(nan? v) (display "SCM_DBL_NAN")]
+          [(infinite? v) (display (if (> v 0)
+                                    "SCM_DBL_POSITIVE_INFINITY"
+                                    "SCM_DBL_NEGATIVE_INFINITY"))]
+          [(-zero? v) (display "-0.0")]
+          [else (format #t "~s" v)]))
   (cond
    [(memq class `(,<s8vector> ,<s16vector> ,<s32vector>))
     (format #t "~d,\n" v)]
@@ -762,29 +805,29 @@
    [(eq? class <s64vector>)
     (print "#if SIZEOF_LONG == 8")
     (format #t "~dl,\n" v)
-    (print "#else if SCM_EMULATE_INT64")
-    (print "#ifdef WORDS_BIGENDIAN")
-    (print "#else  /*!WORDS_BIGENDIAN*/")
-    (print "#endif /*!WORDS_BIGENDIAN*/")
-    (print "#else  /*SIZEOF_LONG == 4 && !SCM_EMULATE_INT64*/")
+    (print "#else  /*SIZEOF_LONG == 4*/")
     (format #t " (((int64_t)~dl << 32)|~dl),\n"
             (ash v -32) (logand v (- (%expt 2 32) 1)))
-    (print "#endif /*SIZEOF_LONG == 4 && !SCM_EMULATE_INT64*/")]
+    (print "#endif /*SIZEOF_LONG == 4*/")]
    [(eq? class <u64vector>)
     (print "#if SIZEOF_LONG == 8")
     (format #t "~dlu,\n" v)
-    (print "#else if SCM_EMULATE_INT64")
-    (print "#ifdef WORDS_BIGENDIAN")
-    (print "#else  /*!WORDS_BIGENDIAN*/")
-    (print "#endif /*!WORDS_BIGENDIAN*/")
-    (print "#else  /*SIZEOF_LONG == 4 && !SCM_EMULATE_INT64*/")
+    (print "#else  /*SIZEOF_LONG == 4*/")
     (format #t " (((int64_t)~dlu << 32)|~dlu),\n"
             (ash v -32) (logand v (- (%expt 2 32) 1)))
-    (print "#endif /*SIZEOF_LONG == 4 && !SCM_EMULATE_INT64*/")]
-   [(eq? class <f16vector>)
-    (error "gauche.cgen.literal: literal f16vector isn't supported yet")]
-   [(eq? class <f32vector>) (format #t "~sf,\n" v)]
-   [(eq? class <f64vector>) (format #t "~s,\n" v)]
+    (print "#endif /*SIZEOF_LONG == 4*/")]
+   [(eq? class <f16vector>) (pr-half-float v) (print ",")]
+   [(eq? class <f32vector>) (pr-float v) (print ",")]
+   [(eq? class <f64vector>) (pr-double v) (print ",")]
+   [(eq? class <c32vector>)
+    (display "{ ") (pr-half-float (real-part v))
+    (display ", ") (pr-half-float (imag-part v)) (print "},")]
+   [(eq? class <c64vector>)
+    (pr-float (real-part v)) (display " + ")
+    (pr-float (imag-part v)) (print " * _Complex_I,")]
+   [(eq? class <c128vector>)
+    (pr-double (real-part v)) (display " + ")
+    (pr-double (imag-part v)) (print " * _Complex_I,")]
    ))
 
 ;; char-set -----------------------------------------------------
@@ -822,6 +865,147 @@
             (if (~ self'case-fold?) "SCM_REGEXP_CASE_FOLD" "0")))
   (static (self) #f))
 
+;; class --------------------------------------------------------
+
+;; We can only dump classes that has globally bound name.
+(define-cgen-literal <cgen-scheme-class> <class>
+  ((defined-module :init-keyword :defined-module)
+   (name :init-keyword :name))
+  (make (value)
+    (when (null? (~ value'defined-modules))
+      (error "Can't precompile a class that doesn't have global binding:"
+             value))
+    (assume (module? (car (~ value'defined-modules))))
+    (assume (symbol? (class-name value)))
+    (make <cgen-scheme-class> :value value
+          :c-name (cgen-allocate-static-datum)
+          :name (cgen-literal (class-name value))
+          :defined-module (cgen-literal (car (~ value'defined-modules)))))
+  (init (self)
+    (let ([m (gensym "mod")] [n (gensym "name")])
+      (format #t "{\n  ScmObj ~a = SCM_OBJ(~a); ScmObj ~a = ~a;\n"
+              m (cgen-c-name (~ self'defined-module))
+              n (cgen-c-name (~ self'name)))
+      (format #t "  SCM_ASSERT(SCM_MODULEP(~a));\n" m)
+      (format #t "  SCM_ASSERT(SCM_SYMBOLP(~a));\n" n)
+      (format #t "  ~a = Scm_GlobalVariableRef(SCM_MODULE(~a), SCM_SYMBOL(~a),0);\n"
+              (cgen-c-name self) m n)
+      (format #t "}\n")))
+  (static (self) #f))
+
+;; descriptive type ---------------------------------------------
+
+(define-cgen-literal <cgen-scheme-proxy-type> <proxy-type>
+  ((id :init-keyword :id))
+  (make (value)
+    (make <cgen-scheme-proxy-type> :value value
+          :c-name (cgen-allocate-static-datum)
+          :id ($ cgen-literal
+                 $ (with-module gauche.internal proxy-type-id) value)))
+  (init (self)
+        (format #t "  ~a = Scm_MakeProxyType(SCM_IDENTIFIER(~a), NULL);"
+                (cgen-c-name self)
+                (cgen-c-name (~ self'id))))
+  (static (self) #f))
+
+(define-cgen-literal <cgen-scheme-descriptive-type> <descriptive-type>
+  ((ctor :init-keyword :ctor)
+   (args :init-keyword :args))
+  (make (value)
+    (make <cgen-scheme-descriptive-type> :value value
+          :c-name (cgen-allocate-static-datum)
+          :ctor (cgen-literal (class-of value))
+          :args ($ cgen-literal
+                   $ (with-module gauche.internal deconstruct-type) value)))
+  (init (self)
+        (format #t "  ~a = Scm_ConstructType(~a, ~a);"
+                (cgen-c-name self)
+                (cgen-c-name (~ self'ctor))
+                (cgen-c-name (~ self'args))))
+  (static (self) #f))
+
+(define-cgen-literal <cgen-scheme-native-type> <native-type>
+  ((name :init-keyword :name))
+  (make (value)
+    (make <cgen-scheme-native-type> :value value
+          :c-name (cgen-allocate-static-datum)
+          :name (cgen-literal (~ value'name))))
+  (init (self)
+        (format #t "  ~a = Scm_GlobalVariableRef(Scm_GaucheModule(), \
+                                     SCM_SYMBOL(~a), 0);"
+                (cgen-c-name self) (cgen-c-name (~ self'name)))
+        (format #t "  if (SCM_UNBOUNDP(~a)) \
+                             Scm_Error(\"Invalid native type: %S\", ~a);"
+                (cgen-c-name self) (cgen-c-name (~ self'name))))
+  (static (self) #f))
+
+;; module -------------------------------------------------------
+
+;; We recover modules at runtime.  Anonymous modules can be recovered
+;; if they are used as a proxy of a real module (via <tmodule>).
+(define-cgen-literal <cgen-module> <module>
+  ((name  :init-keyword :name)
+   (name-literal :init-keyword :name-literal))
+  (make (value)
+    (if-let1 n (predefined-module value)
+      (make <cgen-module> :value value
+            :c-name n
+            :name (~ value'name))
+      (let1 name (or (~ value'name)
+                     (any (^[tm] (and (eq? (~ tm'module) value) (~ tm'name)))
+                          (all-tmodules))
+                     (error "Cannot emit literal for anonymous module:" value))
+        (make <cgen-module> :value value
+              :c-name (cgen-allocate-static-datum)
+              :name name :name-literal (cgen-literal name)))))
+  (init (self)
+    (unless (predefined-module (~ self'value))
+      ;; Module literals are emitted only related with macro hygiene.  In that
+      ;; case, the module is created without loading its actual body, so
+      ;; we mark it as 'placeholding'.  The placeholding module won't show
+      ;; up in (all-modules).  Once (define-module ...) is evaluated by loading
+      ;; the module, it becomes 'real' module.
+      (format #t "  ~a = SCM_OBJ(Scm_FindModule(SCM_SYMBOL(~a), \
+                         SCM_FIND_MODULE_CREATE|SCM_FIND_MODULE_PLACEHOLDING)); \
+                  /* module ~a */\n"
+              (~ self'c-name)
+              (cgen-cexpr (~ self'name-literal))
+              (~ self'name))))
+  (static (self) (boolean (predefined-module (~ self'value))))
+  )
+
+(define (predefined-module module)
+  (assq-ref `((,(find-module 'null)   . "Scm_NullModule()")
+              (,(find-module 'scheme) . "Scm_SchemeModule()")
+              (,(find-module 'gauche) . "Scm_GaucheModule()")
+              (,(find-module 'gauche.internal) . "Scm_GaucheInternalModule()")
+              )
+            module))
+
+;; identiifer ---------------------------------------------------
+
+(define-cgen-literal <cgen-scheme-identifier> <identifier>
+  ((id-name        :init-keyword :id-name)
+   (module-literal :init-keyword :module-literal))
+  (make (value)
+    (unless ((with-module gauche.internal identifier-toplevel?) value)
+      (error "identifier with compiler environment can't be compiled" value))
+    (make <cgen-scheme-identifier> :value value
+          :c-name (cgen-allocate-static-datum)
+          :id-name (cgen-literal (unwrap-syntax value))
+          :module-literal (cgen-literal (~ value'module))))
+  (init (self)
+    (let ([name (cgen-cexpr (~ self'id-name))]
+          [cname (~ self'c-name)])
+      (format #t "  ~a = Scm_MakeIdentifier(~a, SCM_MODULE(~a), \
+                                            SCM_NIL); /* ~a#~a */\n"
+              cname name (cgen-cexpr (~ self'module-literal))
+              (cgen-safe-comment (~ self'module-literal'name))
+              (cgen-safe-comment (~ self'id-name'value)))))
+  (static (self) #f)
+  )
+
+
 ;;---------------------------------------------------------------
 ;; Inferring literal handlers.
 ;;  The code generator does not know how to generate literals for
@@ -875,4 +1059,3 @@
 (define cgen-unique-name
   (let1 counter 0
     (^[] (format "~5,'0d" (inc! counter)))))
-
