@@ -1,7 +1,7 @@
 ;;;
 ;;; gauche.interactive.editable-reader
 ;;;
-;;;   Copyright (c) 2016-2021  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2016-2022  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -40,9 +40,10 @@
   (export make-editable-reader))
 (select-module gauche.interactive.editable-reader)
 
-;; Delay loading line-edit in case editable console isn't available.
+;; Delay loading some modules in case editable console isn't available.
 (autoload text.line-edit <line-edit-context> read-line/edit
           load-line-edit-history save-line-edit-history)
+(autoload gauche.interactive.completion list-completions)
 
 ;; Internal API, to be used by gauche.interactive.
 ;; Because of toplevel commands, we can't just provide alternative 'read'
@@ -54,10 +55,14 @@
 ;; to allow other console (e.g. over pty).
 (define (make-editable-reader get-prompt-string hist-file)
   (if-let1 console (make-default-console :if-not-available #f)
+    ;; We wrap list-completions with lambda to delay loading completion module
     (let ([ctx (make <line-edit-context>
                  :console console
                  :prompt (^[] (display (get-prompt-string)))
-                 :input-continues (^s (not (%input-complete? s))))]
+                 :input-continues (^s (not (%input-complete? s)))
+                 :completion-word-constituent? char-word-constituent?
+                 :completion-lister (^[word gbuf start end]
+                                      (list-completions word gbuf start end)))]
           [buffer (open-input-string "")])
       (define (read-1 reader)
         (rec (try)
