@@ -1,7 +1,7 @@
 ;;;
 ;;; interactive/toplevel.scm - toplevel commands
 ;;;
-;;;   Copyright (c) 2015-2021  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2015-2022  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -53,7 +53,7 @@
   (use srfi-13)
   (use util.match)
   (use data.trie)
-  (export handle-toplevel-command)
+  (export handle-toplevel-command toplevel-command-matches)
   )
 (select-module gauche.interactive.toplevel)
 
@@ -77,6 +77,12 @@
              `((,k . ,v))
              (sort (trie-common-prefix t k) string<? car))))))
 
+;; For gauche.interactive.completion
+;; Returns a lis tof commands that shares the given prefix
+(define (toplevel-command-matches key)
+  ($ atomic *toplevel-commands*
+     (^t (sort (map car (trie-common-prefix t key))))))
+
 ;; Returns [((key ...) help-string)]
 (define (toplevel-command-keys)
   ($ atomic *toplevel-commands*
