@@ -1,7 +1,7 @@
 ;;;
 ;;; library utilities - to be autoloaded.
 ;;;
-;;;   Copyright (c) 2003-2021  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2003-2022  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -43,10 +43,9 @@
 ;;  proc takes the matched module/library name, full pathname, and seed value.
 ;;  This can be more involved once customized module mapping system
 ;;  is introduced; for now, we simply apply the default mapping rule.
-
 (define (library-fold pattern proc seed
                       :key (paths *load-path*) (allow-duplicates? #f)
-                      (strict? #t))
+                           (strict? #t))
 
   (define search-module?
     (cond [(string? pattern) #f]
@@ -55,74 +54,66 @@
 
   (define seen '())
 
-  ;; pats - list of pattern components splitted by '.' or '/'
-  ;; prefix - one of load paths, e.g. /usr/share/gauche/site/lib/
-  ;; file - path components after prefix, e.g. gauche/mop
-  ;; base - the last component of file, e.g. mop
-  (define (search pats prefix file base seed)
-    (let1 path (topath prefix file)
-      (cond [(and (not (null? (cdr pats)))
-                  (match? (car pats) base)
-                  (file-is-directory? path))
-             (fold (^[subfile seed]
-                     (search (cdr pats) prefix #"~|file|/~|subfile|"
-                             subfile seed))
-                   seed (readdir path))]
-            [(and (null? (cdr pats))
-                  (or (string-suffix? ".scm" base)
-                      (string-suffix? ".sci" base))
-                  (match? (car pats) (string-drop-right base 4))
-                  (file-exists? path))
-             (cond [allow-duplicates?  (proc (ensure path file) path seed)]
-                   [(member file seen) seed]
-                   [(ensure path file)
-                    => (^[mod] (push! seen file) (proc mod path seed))]
-                   [else seed])]
-            [else seed])))
+  (define (decompose path)
+    (let1 cs (string-split path #[/\\])
+      (if (null? cs) '(".") cs)))
 
-  (define (match? pat component)
-    (let1 rx (module-glob-pattern->regexp pat)
-      (cond [(rx component) => (^m (m 0))]
-            [else #f])))
-  (define (ensure path file)
+  (define (compose components)
+    (string-join components (cond-expand
+                             [gauche.os.windows "\\"]
+                             [else "/"])))
+
+  ;; For module path (e.g. "util/match" for util.match), we always use '/'.
+  (define (ensure path partial)
+    (define modpath (string-drop-right (string-join partial "/") 4))
     (if search-module?
-      (let1 modname (path->module-name (string-drop-right file 4))
+      (let1 modname (path->module-name modpath)
         (and (or (not strict?)
-                 (library-has-module? path modname))
+                 (library-has-module? (compose path) modname))
              modname))
-      (string-drop-right file 4)))
+      modpath))
 
-  ;; main body
-  (let1 pats (if search-module?
-               (string-split (x->string pattern) #\.)
-               (string-split pattern #\/))
+  (define (fold-dirs proc seed dirpath regexp)
+    (fold (^[elt seed]
+            (or (and-let* ([ (regexp elt) ]
+                           [path (append dirpath `(,elt))]
+                           [ (file-is-directory? (compose path)) ])
+                  (proc path seed))
+                seed))
+          seed (readdir (compose dirpath))))
+
+  (define (fold-leaf proc seed prefix dirpath regexp)
+    (fold (^[elt seed]
+            (or (and-let* ([m (#/(.*)\.sc[im]$/ elt) ]
+                           [base (m 1)]
+                           [ (regexp base) ]
+                           [path (append dirpath `(,elt))]
+                           [file (drop path (length prefix))]
+                           [key (ensure path file)])
+                  (if (and (not allow-duplicates?) (member key seen))
+                    seed
+                    (begin (push! seen key) (proc key (compose path) seed))))
+                seed))
+          seed (readdir (compose dirpath))))
+
+  (define (make-folder prefix)
+    (^[proc seed parent regexp non-leaf?]
+      ;; We always search to the leaf nodes, so REGEXP never be 'dir.
+      (if non-leaf?
+        (fold-dirs proc seed parent regexp)
+        (fold-leaf proc seed prefix parent regexp))))
+
+  (let ([pattern (if search-module?
+                   (module-name->path pattern)
+                   pattern)])
     (fold (^[prefix seed]
-            (fold (^[file seed] (search pats prefix file file seed))
-                  seed (readdir prefix)))
-          seed paths))
-
-  #|
-  (define (get-relative prefix path separ)
-  (string-trim (string-drop path (string-length prefix)) #[/\\]))
-
-  (define (picker prefix separ path seed)
-  (cond [(not (string-suffix? ".scm" path)) seed]
-  [(and-let* ([file #?=(get-relative prefix path separ)]
-  [ (or allow-duplicates? (not (member file seen))) ]
-  [mod (ensure path file)])
-  (push! seen file)
-  (proc mod path seed))]
-  [else seed]))
-
-  (receive (pat sep) (if search-module?
-  (values (x->string pattern) #[.])
-  (values pattern #[/]))
-  (fold (lambda (prefix seed)
-  (glob-fold pat (cut picker prefix sep <> <>) seed
-  :separator sep :current prefix))
-  seed paths))
-  |#
-  )
+            (let1 prefix ($ decompose
+                            $ sys-normalize-pathname prefix :canonicalize #t)
+              (glob-fold pattern proc seed
+                         :folder (make-folder prefix)
+                         :prefix prefix
+                         :sorter #f)))
+          seed paths)))
 
 ;; Just check existence of library.
 (define (library-exists? mod/path :key (force-search? #f)
