@@ -1,7 +1,7 @@
 ;;;
 ;;; cgi.scm - CGI utility
 ;;;
-;;;   Copyright (c) 2000-2019  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2000-2022  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -38,12 +38,11 @@
 ;; the httpd.
 
 (define-module www.cgi
-  (use srfi-1)
+  (use scheme.list)
   (use srfi-13)
   (use rfc.uri)
   (use rfc.mime)
   (use rfc.822)
-  (use gauche.parameter)
   (use gauche.charconv)
   (use gauche.uvector)
   (use text.tree)
@@ -430,7 +429,7 @@
     ,(html:html
       (html:head (html:title "Error"))
       (html:body (html:h1 "Error")
-                 (html:p (html-escape-string (slot-ref e 'message)))))))
+                 (html:p (html-escape-string (condition-message e e)))))))
 
 (define (cgi-default-output tree)
   (if (ces-equivalent? (gauche-character-encoding)
@@ -443,4 +442,3 @@
      (current-output-port)
      (cut write-tree tree)
      :encoding (cgi-output-character-encoding))))
-
